import numpy as np


def crossover_1point(self):
    chrom, n_pop, len_chrom = self.chrom, self.n_pop, self.len_chrom
    # 每次前后两个染色体进行交叉
    for i in range(0, n_pop, 2):
        # 染色体长度中，随便选择一个点
        n = np.random.randint(0, self.len_chrom)
        # 交叉合并
        seg1, seg2 = self.chrom[i, n:].copy(), self.chrom[i + 1, n:].copy()
        self.chrom[i, n:], self.chrom[i + 1, n:] = seg2, seg1
    return self.chrom


def crossover_2point(self):
    chrom, n_pop, len_chrom = self.chrom, self.n_pop, self.len_chrom
    for i in range(0, n_pop, 2):
        # 两个点，
        n1, n2 = np.random.randint(0, self.len_chrom, 2)
        if n1 > n2:
            n1, n2 = n2, n1
        # 将两个点内的染色体交换
        seg1, seg2 = self.chrom[i, n1:n2].copy(), self.chrom[i + 1, n1:n2].copy()
        self.chrom[i, n1:n2], self.chrom[i + 1, n1:n2] = seg2, seg1
    return self.chrom



import numpy as np
import pandas as pd

from scipy.optimize import minimize


def smc(corr_mtx):
    """
    主因子法：第一种估计h的方法
    :param corr_mtx:
    :param sort:
    :return:
    """
    corr_inv = np.linalg.inv(corr_mtx)
    # r_ii = R^-1
    # std = 1/r_ii
    # h = 1 - std
    smc = 1 - 1 / np.diag(corr_inv)
    return smc


def corr(x: np.array):
    x = (x - x.mean(0)) / x.std(0)
    return np.cov(x, rowvar=False, ddof=0)


class Factor:
    def __init__(self,
                 n_factors=3,
                 rotaion='promax',
                 method='ml',
                 use_smc=True,
                 bounds=(0.005, 1)):
        """
            use_smc 是否使用平方多重相关作为因子分析的起始猜测
        """
        self.bounds = bounds
        self.use_smc = use_smc
        self.method = method
        self.rotaion = rotaion
        self.n_factors = n_factors

        self.loadings_ = None
        self.rotation_matrix_ = None

    def fit(self, X):
        if isinstance(X, pd.DataFrame):
            X = X.values.copy()
        else:
            X = X.copy()

        # 获取相关矩阵
        corr_mtx = corr(X)
        self.std_ = np.std(X, axis=0)
        self.mean_ = np.mean(X, axis=0)

        if self.method == 'principal':
            loadings = self._fit_principal(corr_mtx)
        else:
            loadings = self._fit_ml(corr_mtx)

        self.loadings_ = loadings
        return self.loadings_

    def _fit_principal(self, corr_mtx):

        pass

    def _fit_ml(self, corr_mtx):
        """
        估计值的预处理
        :param corr_mtx:
        :return:
        """
        # 估计误差的值
        if self.use_smc:
            smc_mtx = smc(corr_mtx)
            # 随机初始化ψ_0值
            start = (np.diag(corr_mtx) - smc_mtx.T).squeeze()
        else:
            # 随机值设置0.5
            start = [0.5 for _ in range(corr_mtx.shape[0])]

        # this must be a list passed to `minimize()`
        bounds = self.bounds

        # 将随机估计的误差值，通过 minimize进最小化(梯度下降，巨呃度下降，牛顿下降)
        objective = self._fit_ml_objective
        res = minimize(objective,
                       start,
                       method='CG',
                       bounds=bounds,
                       options={'maxiter': 1000},
                       args=(corr_mtx, self.n_factors))

        # 求因子载荷，通过极小误差
        return self._normalize_ml(res.x, corr_mtx, self.n_factors)

    @staticmethod
    def _fit_ml_objective(psi, corr_mtx, n_factors):
        """
        求估计误差极小值
        :param psi: 主因子法估计的误差值
        :param corr_mtx:
        :param n_factors:
        :return:
        """
        std_0_ = np.diag(1 / np.sqrt(psi))
        std_s_std = np.dot(np.dot(std_0_, corr_mtx), std_0_)

        feature, _ = np.linalg.eigh(std_s_std)
        feature = feature[::-1][n_factors:]

        error = -(np.sum(np.log(feature) - feature) - n_factors + corr_mtx.shape[0])
        return error

    def _normalize_ml(self, x, corr_mtx, n_factors):
        """
        通过ψ_0 求 Λ_0 , Λ_0 -> ψ_1
        :param x:
        :param corr_mtx:
        :param n_factors:
        :return:
        """
        std_0_ = np.diag(1 / np.sqrt(x))
        std_s_std = np.dot(np.dot(std_0_, corr_mtx), std_0_)

        feature, vector = np.linalg.eigh(std_s_std)
        feature = feature[::-1][:n_factors]
        vector = vector[:, ::-1][:, :n_factors]

        feature = np.maximum(feature - 1, 0)

        loadings = np.dot(vector, np.diag(np.sqrt(feature)))

        return np.dot(np.diag(np.sqrt(x)), loadings)


if __name__ == '__main__':
    data = pd.read_excel('../../data/pathanalysis.xlsx', index_col=0)
    fa = Factor(n_factors=1)
    loading = fa.fit(data)
    print(loading)



import numpy as np


def selection_roulette_1(self):
    '''
    普通版本轮盘赌
    :param self:
    :return:
    '''
    FitV = self.FitV
    FitV = FitV - FitV.min() + 1e-10
    sel_prob = FitV / FitV.sum()
    # 有放回抽取,因为有放回，代表同一个染色体可以被抽到多次进行复制
    sel_index = np.random.choice(range(self.n_pop), size=self.n_pop, p=sel_prob)
    self.chrom = self.chrom[sel_index, :]
    return self.chrom


def selection_roulette_2(self):
    '''
    标准化那版本轮盘赌
    :param self:
    :return:
    '''
    FitV = self.FitV
    FitV = (FitV - FitV.min()) / (FitV.max() - FitV.min() + 1e-10) + 0.2
    # the worst one should still has a chance to be selected
    sel_prob = FitV / FitV.sum()
    # 有放回抽取
    sel_index = np.random.choice(range(self.n_pop), size=self.n_pop, p=sel_prob)
    self.chrom = self.chrom[sel_index, :]
    return self.chrom



import numpy as np
from ._base import proBase
from abc import ABCMeta, abstractmethod
from .tool import func_transformer
from .operators import crossover, mutation, ranking, selection


class GeneticAlgorithmBase(proBase, metaclass=ABCMeta):
    def __init__(self,
                 fit_func,
                 n_feature,
                 n_pop=100,
                 n_gen=1000,
                 prob_mut=0.001):
        self.fit_func = func_transformer(fit_func)
        if isinstance(n_pop, int):
            self.n_pop = n_pop
        else:
            self.n_pop = round(n_pop)
        self.prob_mut = prob_mut
        self.n_gen = n_gen
        self.n_feature = n_feature

        self.chrom = None
        self.X = None  # 染色体
        self.Y_raw = None  # shape = (size_pop,) , value is f(x)
        self.Y = None  # 目标函数值
        self.fit_v = None  # shape = (size_pop,)

        self.generation_best_X = []
        self.generation_best_Y = []

        self.all_history_Y = []
        self.all_history_FitV = []

        self.best_x, self.best_y = None, None

    # 增加限制(目前暂不提供)
    def x2y(self):
        """
        1. 染色体进入目标函数运行
        2. 添加目标函数结果限制，
        :return:
        """
        self.Y_raw = self.fit_func(self.X)
        # if not self.has_constraint:
        self.Y = self.Y_raw
        # else:
        #     # constraint
        #     penalty_eq = np.array([np.sum(np.abs([c_i(x) for c_i in self.constraint_eq])) for x in self.X])
        #     penalty_ueq = np.array([np.sum(np.abs([max(0, c_i(x)) for c_i in self.constraint_ueq])) for x in self.X])
        #     self.Y = self.Y_raw + 1e5 * penalty_eq + 1e5 * penalty_ueq
        return self.Y

    @abstractmethod
    def chrom2x(self, chrom):
        pass

    @abstractmethod
    def ranking(self):
        pass

    @abstractmethod
    def selection(self):
        pass

    @abstractmethod
    def crossover(self):
        pass

    @abstractmethod
    def mutation(self):
        pass

    def run(self, n_gen=None):
        # 可以在run的时候重新设置存活代数
        self.n_gen = n_gen or self.n_gen
        for i in range(self.n_gen):
            # 二进制转实数
            self.X = self.chrom2x(self.chrom)
            # 目前没有限制条件,目标函数值
            self.Y = self.x2y()
            self.ranking()  # == self.Y
            self.selection()
            self.crossover()
            self.mutation()
            # 新一代的孩子染色体已经生成好，下面保存上一代的最好值
            # 当前这代最好目标函数的索引
            generation_best_index = self.FitV.argmax()
            # 当前这代最好的染色体实数
            self.generation_best_X.append(self.X[generation_best_index, :])
            # 当前这代最好的染色体目标函数值
            self.generation_best_Y.append(self.Y[generation_best_index])
            # 历史目标函数值
            self.all_history_Y.append(self.Y)
            # 历史繁殖了的染色体
            self.all_history_FitV.append(self.FitV)

        # 所有最好的目标函数中最小值的索引
        global_best_index = np.array(self.generation_best_Y).argmin()
        # 所有最好的目标函数中最小值的染色体
        self.best_x = self.generation_best_X[global_best_index]
        # 最小值染色体的目标函数值
        self.best_y = self.fit_func(np.array([self.best_x]))
        return self.best_x, self.best_y

    fit = run


class GA(GeneticAlgorithmBase):
    def __init__(self,
                 fit_func,
                 n_feature,
                 n_pop=100,
                 n_gen=1000,
                 prob_mut=0.001,
                 lb=[-1, -1, -1],
                 ub=[1, 1, 1],
                 precision=1e-7):
        """

        :param fit_func: 目标函数
        :param n_feature: 袋鼠特征个数
        :param n_pop: 每一代多少个袋鼠
        :param n_gen: 一个多少代
        :param prob_mut: 变异几率
        :param prob_cross: 交叉几率
        :param lb: 随机初始特征下限值
        :param ub: 随机初始特征上限值
        :param precision: func中每个变量的精度
        """
        super().__init__(fit_func, n_feature, n_pop, n_gen, prob_mut)

        self.lb = np.array(lb) if isinstance(lb, list) else lb
        self.ub = np.array(ub) if isinstance(ub, list) else ub
        # 二进制的精度
        # 精度
        self.precision = np.array(precision) * np.ones(n_feature)

        # 根据区间，生成染色体长
        # 精度：几位, 6位
        # (ub - lb) = ?
        # 2**(?-1) < ?*10^6 < 2**(?)
        # 两个方法都可以
        # Lind_raw = np.log2((self.ub - self.lb) * self.precision + 1)
        Lind_raw = np.log2((self.ub - self.lb) / self.precision + 1)
        self.Lind = np.ceil(Lind_raw).astype(int)

        # self.int_mode_ = (self.precision % 1 == 0) & (Lind_raw % 1 != 0)
        # self.int_mode = np.any(self.int_mode_)
        #
        # if self.int_mode:
        #     self.ub_extend = np.where(self.int_mode_
        #                               , self.lb + (np.exp2(self.Lind) - 1) * self.precision
        #                               , self.ub)

        # 染色体的长度 = 所有精度的总长度
        self.len_chrom = sum(self.Lind)

        self.creata_pop_chrom()

    def creata_pop_chrom(self):
        # 生成self.n_pop个二进制染色体，每个染色体长self.len_chrom
        self.chrom = np.random.randint(low=0, high=2, size=(self.n_pop, self.len_chrom))
        return self.chrom

    def gray2rv(self, gray_code):
        # 二进制转十进制
        _, len_gray_code = gray_code.shape
        b = gray_code.cumsum(axis=1) % 2
        mask = np.logspace(start=1, stop=len_gray_code, base=0.5, num=len_gray_code)
        return (b * mask).sum(axis=1) / mask.sum()

    def chrom2x(self, chrom):
        # 获取每个特征的长度
        cumsum_len_segment = self.Lind.cumsum()
        X = np.zeros((self.n_pop, self.n_feature))
        # 将第一个特征值先取出来
        chrom_temp = chrom[:, :cumsum_len_segment[0]]
        # 二进制转成实数
        X[:, 0] = self.gray2rv(chrom_temp)
        # 取出后续的值
        for feat in range(1, len(cumsum_len_segment)):
            chrom_temp = chrom[:, cumsum_len_segment[feat - 1]:cumsum_len_segment[feat]]
            X[:, feat] = self.gray2rv(chrom_temp)

        # if self.int_mode:
        #     X = self.lb + (self.ub_extend - self.lb) * X
        #     X = np.where(X > self.ub, self.ub, X)
        #     # the ub may not obey precision, which is ok.
        #     # for example, if precision=2, lb=0, ub=5, then x can be 5
        # else:
        # 将二进制设置在区内
        X = self.lb + (self.ub - self.lb) * X
        return X

    ranking = ranking.ranking
    selection = selection.selection_roulette_1
    crossover = crossover.crossover_1point
    mutation = mutation.mutation



from abc import ABCMeta, abstractmethod
import types
import warnings
from Statech.spssbase import SpssPro


class proBase(SpssPro, metaclass=ABCMeta):
    def register(self, operator_name, operator, *args, **kwargs):
        def operator_wapper(*wrapper_args):
            return operator(*(wrapper_args + args), **kwargs)

        # setattr() 函数对应函数 getattr()，用于设置属性值，该属性不一定是存在的。
        setattr(self, operator_name, types.MethodType(operator_wapper, self))
        return self

    def fit(self, *args, **kwargs):
        warnings.warn('.fit() will be deprecated in the future. use .run() instead.'
                      , DeprecationWarning)
        return self.run(*args, **kwargs)



class AlgorithmError(RuntimeError):
    """
    算法异常抛出
    如果知道异常来源比如被除数为 0 的话: raise AlgorithmError("A 列数据不能存在 0 值") from ZeroDivisionError
    如果不知道异常来源的话直接: raise AlgorithmError("A 列数据错误")
    """
    pass



import numpy as np
from functools import lru_cache
from types import MethodType, FunctionType
import warnings
import sys
import multiprocessing


def func_transformer(func):
    # getattr() 函数用于返回一个对象属性值。
    # func不包含’mode‘ = ’other'
    mode = getattr(func, 'mode', 'others')
    # 当func包含mode，mode的值只能执行这些类型的模型，
    valid_mode = ('common', 'multithreading', 'multiprocessing', 'vectorization', 'cached', 'others')
    assert mode in valid_mode, 'valid mode should be in ' + str(valid_mode)
    if mode == 'vectorization':
        return func
    elif mode == 'cached':
        # @lru_cache装饰器可以提供缓存的功能,在下次函数接收到相同参数调用时直接返回上一次的结果,用以节约高开销或I/O函数的调用时间
        # maxsize :  None，LRU 特性将被禁用且缓存可无限增长
        # type : True不同类型被分开存放
        @lru_cache(maxsize=None)
        def func_cached(x):
            return func(x)

        def func_warped(X):
            return np.array([func_cached(tuple(x)) for x in X])

        return func_warped
    elif mode == 'multithreading':
        # 是否使用线程池
        from multiprocessing.dummy import Pool as ThreadPool

        pool = ThreadPool()

        def func_transformed(X):
            return np.array(pool.map(func, X))

        return func_transformed
    elif mode == 'multiprocessing':
        # 是否使用多进程
        from multiprocessing import Pool
        pool = Pool()

        def func_transformed(X):
            return np.array(pool.map(func, X))

        return func_transformed
    else:
        # 普遍的方法
        def func_transformed(X):
            return np.array([func(x) for x in X])

        return func_transformed



import numpy as np


def ssr(y_true, y_pred):
    y_mean = np.mean(y_true)
    return np.sum(np.power((y_pred - y_mean), 2))


def sse(y_true, y_pred):
    return np.sum(np.power((y_true - y_pred), 2))


def sst(y_true, y_pred):
    return ssr(y_true, y_pred) + sse(y_true, y_pred)


def mse(y_true, y_pred, n, k):
    """
    :param y_true: 真实值
    :param y_pred: 预测值
    :param n: 自由度
    :param k: 可控自由
    :return:
    """
    return sse(y_true, y_pred) / (n - k - 1)


def mst(y_true, y_pred, n, k):
    return sst(y_true, y_pred) / (n - k - 1)


def msr(y_true, y_pred, k):
    """

    :param y_true: 真实值
    :param y_pred: 预测值
    :param k: 可控自由度
    :return:
    """
    return ssr(y_true, y_pred) / k


def r_square(y_true, y_pred):
    return ssr(y_true, y_pred) / sst(y_true, y_pred)


def adjust_r_square(y_true, y_pred, n, k):
    return 1 - (1 - r_square(y_true, y_pred)) * (n - 1) / (n - k - 1)


def f(y_true, y_pred, n, k):
    return msr(y_true, y_pred, k) / mse(y_true, y_pred, n, k)


def bse(X: np.ndarray, y_true, y_pred, n, k):
    normalized_cov_params = np.linalg.inv(np.dot(X.T, X))
    cov_p = normalized_cov_params * mse(y_true, y_pred, n, k)
    return np.sqrt(np.diag(cov_p))


def t(theta, X: np.ndarray, y_true, y_pred, n, k):
    return theta / bse(X, y_true, y_pred, n, k)


def t_p(theta, X: np.ndarray, t_value, n, k):
    from scipy.stats import t
    return t.sf(abs(t_value), n - k - 1) * 2


def std_coef(coef, X: np.ndarray, y):
    x_std = np.std(X, axis=0)
    y_std = np.std(y)
    return np.append(0, np.multiply(coef, (x_std / y_std)))


def f_change(ssr_old, ssr, mse, gap):
    return ((ssr - ssr_old) / gap) / mse


def f_p(f_change, gap, n, k):
    from scipy.stats import f
    return f.sf(abs(f_change), gap, n - k - 1)



import pandas as pd
import numpy as np
from typing import Union
import itertools
from ._base import LinearModel, RegressorMixin


class LinearRegression(RegressorMixin, LinearModel):
    def fit(self, X, y):
        X, y = self._validate_data(X, y)
        theta = np.linalg.inv(X.T * X) * X.T * y.T
        np_theta = np.array(theta).T.ravel()
        self._set_intercept(np_theta)
        self._set_coef(np_theta)
        return self


class HierarchicalRegression(RegressorMixin, LinearModel):
    def fit(self, X, y, *args):
        layers = args[0]
        k = 0
        for step in layers:
            k += step
            # 回归
            reg = LinearRegression()
            x_constant = reg.add_constant(X.iloc[:, 0:k])
            self._set_multi_model(reg.fit(x_constant, y))
        return self

    def score(self, X: Union[pd.DataFrame, np.ndarray], y: Union[pd.Series, np.ndarray], *args) -> pd.DataFrame:
        nrow, ncol = X.shape
        k, ssr_old = 0, 0
        result = {}
        layers = args[0]
        f_change_value = pd.Series(0, name='F_change').astype(float)
        f_p_value = pd.Series(0, name='F_change_p').astype(float)
        for model, step, j in zip(self.multi_model_, layers, range(ncol)):
            k += step
            # 普通值
            score = model.score(X.iloc[:, 0:k], y)
            y_pred = model.predict(X.iloc[:, 0:k])
            # 计算f增量
            from Statech.metrics import ssr, mse, f_change, f_p
            mse_value = mse(y.values, y_pred, nrow, k)
            ssr_new = ssr(y.values, y_pred)
            f_change_value[0] = f_change(ssr_old, ssr_new, mse_value, step)
            f_p_value[0] = f_p(f_change_value, step, nrow, k)
            ssr_old = ssr_new
            # 结果
            result['layer' + str(j)] = [score[0], pd.concat([score[1], f_change_value, f_p_value], axis=1)]
        return result


class MediatingEffect(RegressorMixin, LinearModel):
    def fit(self, X, y, *args):
        # 设置变量
        x = X
        m = args[0]
        c = None
        if len(args) > 1:
            c = args[1]
            x = pd.concat([c, x], axis=1)

        # x->y
        reg = LinearRegression()
        x_constant = reg.add_constant(x)
        self._set_multi_model(reg.fit(x_constant, y))

        # x->m
        for i in range(m.shape[1]):
            y_m = m.iloc[:, i]
            reg = LinearRegression()
            self._set_multi_model(reg.fit(x_constant, y_m))

        # x_m->y
        x = pd.concat([x, m], axis=1)
        reg = LinearRegression()
        x_constant = reg.add_constant(x)
        self._set_multi_model(reg.fit(x_constant, y))
        return self

    def score(self, X: Union[pd.DataFrame, np.ndarray], y: Union[pd.Series, np.ndarray], *args) -> pd.DataFrame:
        # 设置样本
        x = X
        m = args[0]
        c = None
        if len(args) > 1:
            c = args[1]
            x_c = pd.concat([x, c], axis=1)

        # 获x->y
        x_y_mode = self.multi_model_.pop(0)
        x_y_m_mode = self.multi_model_.pop()

        # x->y
        xcol = x.shape[1]
        x_y_result = x_y_mode.score(x_c, y)[0]
        c_ = x_y_result.iloc[-xcol:, 0].rename('c')

        # x->m
        mcol = m.shape[1]
        a_ = []
        for i in range(mcol):
            y_m = m.iloc[:, i]
            x_m_result = self.multi_model_[i].score(x_c, y_m)[0]
            a_.append(x_m_result.iloc[-xcol:, 0].values)
        cross_a = a_[0]
        for i in range(1, mcol):
            cross_a = list(itertools.chain.from_iterable(zip(cross_a, a_[i])))
        a_ = pd.Series(np.array(cross_a), name='a')

        # x_m->y
        x_c_m = pd.concat([x_c, m], axis=1)
        x_y_m_result = x_y_m_mode.score(x_c_m, y)[0]
        b_ = x_y_m_result.iloc[-mcol:, 0]



import numpy as np


def mutation(self):
    '''
    mutation of 0/1 type chromosome
    faster than `self.Chrom = (mask + self.Chrom) % 2`
    :param self:
    :return:
    '''
    # 生成一个掩码染色体，然后异或之前的染色体
    mask = (np.random.rand(self.n_pop, self.len_chrom) < self.prob_mut)
    self.chrom ^= mask
    return self.chrom



import setuptools

__version__ = '0.0.3'

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="statis-learn",
    python_requires=">=3.5",
    version=__version__,
    author="Jim Chen",
    author_email="axfbh119@gmail.com",
    description="Statistics in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/axfbh/statis-learn",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",  # License
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    platforms=['linux', 'windows', 'macos'],
    install_requires=['numpy', 'scipy', 'pandas'],
)



import cmath

import numpy as np
import pandas as pd
from scipy.stats import f
from sympy.matrices import Matrix as sy_matrix


def matrixPower(A, n):
    v, Q = np.linalg.eig(A)
    V = np.diag(v ** (n))
    T = Q * V * np.linalg.inv(Q)
    # T = Q * V * (Q ** (-1))
    return T.copy()


def multi(lambdas, k):
    r = len(lambdas)
    sum = 1
    for i in range(k, r):
        sum *= (1 - lambdas[i] ** 2)
    return sum


class CCAs:
    def __init__(self, X: pd.DataFrame, y: pd.DataFrame):
        self.X = X
        self.y = y
        self.n = len(X)
        # corr : 数据标准化后的协方差矩阵
        # cov : 数据的协方差矩阵
        self.R = pd.concat([X, y], axis=1).cov()
        r = self.R.to_numpy()
        self.Sxx = np.mat(r[:X.shape[1], :X.shape[1]])
        self.Syy = np.mat(r[X.shape[1]:, X.shape[1]:])
        self.Sxy = np.mat(r[:X.shape[1]:, X.shape[1]:])
        self.Syx = self.Sxy.T
        self.s = min(X.shape[1], y.shape[1])
        self.p = X.shape[1]
        self.q = y.shape[1]

    def fecth_n(self):
        # 正交求解法
        # self.Nb = QRsolve(self.Syy, self.Syx) * QRsolve(self.Sxx, self.Sxy)
        self.Nb = sy_matrix(self.Syy).QRsolve(sy_matrix(self.Syx)) * sy_matrix(self.Sxx).QRsolve(sy_matrix(self.Sxy))

    def fecth_m(self):
        self.M = matrixPower(self.Sxx, -0.5) * self.Sxy * matrixPower(self.Syy, -0.5)

    def eig(self, m):
        # 注意上面矩阵的列向量才是特征向量，l2范数为 1
        return np.linalg.eig(m)

    def fetch_a_b_fv(self):
        b_feature, b_vector = self.eig(np.mat(self.Nb).astype(np.float64))
        a_feature, a_vector = b_feature, np.mat(sy_matrix(self.Sxx).QRsolve(sy_matrix(self.Sxy))).astype(
            np.float64) * b_vector
        a_vector = (a_vector[:, a_feature.argsort()[::-1]])[:, :self.s]
        b_vector = (b_vector[:, a_feature.argsort()[::-1]])[:, :self.s]
        a_feature = np.sqrt((a_feature[a_feature.argsort()[::-1]])[:self.s])
        b_feature = a_feature
        self.lambdas = b_feature
        canvarx = np.mat(self.X) * a_vector
        canvary = np.mat(self.y) * b_vector
        sdx = np.std(canvarx, axis=0)
        sdy = np.std(canvary, axis=0)
        canvarx /= sdx
        canvary /= sdy
        a_vector /= sdx
        b_vector /= sdy
        xstructcorr = np.corrcoef(self.X, canvarx, False)[:self.p, self.p:]
        ystructcorr = np.corrcoef(self.y, canvary, False)[:self.q, self.q:]
        xcrosscorr = np.corrcoef(self.X, canvary, False)[:self.p, self.p:]
        ycrosscorr = np.corrcoef(self.y, canvarx, False)[:self.q, self.q:]

        xstructcorrsp = np.sum(xstructcorr ** 2, axis=1) / self.p
        ystructcorrsp = np.sum(ystructcorr ** 2, axis=1) / self.q
        xcrosscorrsp = np.sum(xcrosscorr ** 2, axis=1) / self.p
        ycrosscorrsp = np.sum(ycrosscorr ** 2, axis=1) / self.p
        return self.lambdas, a_vector, b_vector, xstructcorr, ystructcorr, xcrosscorr, ycrosscorr, xstructcorrsp, ystructcorrsp, xcrosscorrsp, ycrosscorrsp

    def fetch_a_b_svd(self):
        u, s, v = np.linalg.svd(self.M)
        A = matrixPower(self.Sxx, -0.5) * u
        B = matrixPower(self.Syy, -0.5) * v.T
        self.lambdas = s
        A = A[:, :self.s]
        B = B[:, :self.s]
        canvarx = np.mat(self.X) * A
        canvary = np.mat(self.y) * B
        sdx = np.std(canvarx, axis=0)
        sdy = np.std(canvary, axis=0)
        canvarx /= sdx
        canvary /= sdy
        xstructcorr = np.corrcoef(self.X, canvarx, False)[:self.p, self.p:]
        ystructcorr = np.corrcoef(self.y, canvary, False)[:self.q, self.q:self.q + self.s]
        xcrosscorr = np.corrcoef(self.X, canvary, False)[:self.p, self.p:self.p + self.s]
        ycrosscorr = np.corrcoef(self.y, canvarx, False)[:self.q, self.q:]
        xstructcorrsp = np.sum(xstructcorr ** 2, axis=1) / self.p
        ystructcorrsp = np.sum(ystructcorr ** 2, axis=1) / self.q
        xcrosscorrsp = np.sum(xcrosscorr ** 2, axis=1) / self.p
        ycrosscorrsp = np.sum(ycrosscorr ** 2, axis=1) / self.p
        return self.lambdas, A, B, xstructcorr, ystructcorr, xcrosscorr, ycrosscorr, xstructcorrsp, ystructcorrsp, xcrosscorrsp, ycrosscorrsp

    def wilks(self):
        w_contians = []
        df1 = []
        df2 = []
        f_contians = []
        p_contians = []
        for k in range(0, self.s):
            v = multi(self.lambdas, k)
            r = (self.n - self.s - 1) - ((abs(self.p - self.q) + 1) / 2)
            ndf = (self.p - k) * (self.q - k)
            u = (ndf - 2) / 4
            xx = ((self.p - k) ** 2 + (self.q - k) ** 2) - 5
            t = cmath.sqrt(((self.p - k) ** 2 * (self.q - k) ** 2 - 4) / xx if xx > 0 else 1)
            iv = v ** (1 / t)
            ddf = (r * t) - (2 * u)
            fstat = ((1 - iv) / iv) * (ddf / ndf)
            p = f.sf(abs(fstat), ndf, round(ddf.real))
            w_contians.append(v.real)
            df1.append(ndf.real)
            df2.append(ddf.real)
            f_contians.append(fstat.real)
            p_contians.append(p.real)
        return w_contians, df1, df2, f_contians, p_contians


def main(X: pd.DataFrame, y: pd.DataFrame):
    cca = CCAs(X, y)
    cca.fecth_n()
    lambdas, a, b, \
    xstructcorr, ystructcorr, xcrosscorr, \
    ycrosscorr, xstructcorrsp, ystructcorrsp, \
    xcrosscorrsp, ycrosscorrsp = cca.fetch_a_b_fv()

    # cca.fecth_m()
    # lambdas, a, b, \
    # xstructcorr, ystructcorr, xcrosscorr, \
    # ycrosscorr, xstructcorrsp, ystructcorrsp, \
    # xcrosscorrsp, ycrosscorrsp = cca.fetch_a_b_svd()

    s = len(lambdas)
    wilks, ndf, ddf, fstat, p = cca.wilks()
    result1 = pd.DataFrame(np.array([lambdas, wilks, ndf, ddf, fstat, p]).T,
                           columns=['相关性', 'Wilks 统计量', '模型自由的', '误差自由度', 'F', 'Pr > F'])

    result2 = pd.DataFrame(a[:, :s], columns=['典型变量X' + str(i) for i in range(s)])
    result3 = pd.DataFrame(b[:, :s], columns=['典型变量Y' + str(i) for i in range(s)])
    xstructcorr = pd.DataFrame(xstructcorr, columns=['典型变量X' + str(i) for i in range(s)])
    xcrosscorr = pd.DataFrame(xcrosscorr, columns=['交叉典型变量X' + str(i) for i in range(s)])
    ystructcorr = pd.DataFrame(ystructcorr, columns=['典型变量Y' + str(i) for i in range(s)])
    ycrosscorr = pd.DataFrame(ycrosscorr, columns=['交叉典型变量Y' + str(i) for i in range(s)])
    result4 = pd.DataFrame(pd.concat([xstructcorr, xcrosscorr], axis=1))
    result5 = pd.DataFrame(pd.concat([ystructcorr, ycrosscorr], axis=1))

    xstructcorrsp = pd.Series(xstructcorrsp)
    ystructcorrsp = pd.Series(ystructcorrsp)

    xcrosscorrsp = pd.Series(xcrosscorrsp)
    ycrosscorrsp = pd.Series(ycrosscorrsp)

    result6 = pd.concat([xstructcorrsp, ycrosscorrsp], axis=0, ignore_index=True)
    result7 = pd.concat([ystructcorrsp, xcrosscorrsp], axis=0, ignore_index=True)
    result8 = pd.concat([result6, result7], axis=1)
    result8.rename(columns={0: '变量X', 1: '变量Y'}, inplace=True)
    print(result1)
    print(result2)
    print(result3)
    print(result4)
    print(result5)
    print(result8)


def normalization(data: pd.DataFrame):
    return (data - data.mean())


if __name__ == '__main__':
    # data = pd.read_excel('../../data/CAA.xlsx')
    # X = data[['x1反复横向跳', 'x2纵跳', 'x3背力', 'x4握力', 'x5台阶试验', 'x6立定体前屈', 'x7俯卧上体后仰']]
    # Y = data[['y1 50米跑', 'y2跳远', 'y3投球', 'y4引体向上', 'y5耐力跑']]
    data = pd.read_excel('../../data/dianxing.xlsx')
    X = data[data.columns[:3]]
    Y = data[data.columns[3:]]
    X = normalization(X)
    Y = normalization(Y)
    main(X, Y)



from abc import ABCMeta, abstractmethod
import pandas as pd
import numpy as np
from Statech.metrics import *
from typing import Union

from Statech.spssbase import SpssPro
from Statech.pojo.errors import AlgorithmError


class LinearModel(SpssPro, metaclass=ABCMeta):
    """Base class for Linear Models"""

    def __init__(self):
        self.multi_model_ = []

    @abstractmethod
    def fit(self, X, y, *args):
        """Fit model."""
        pass

    def _validate_data(self, X, y=None) -> np.mat:
        """
        数据类型检验转换，
        :param X:
        :param y:
        :return: 矩阵X,矩阵y
        """
        if not isinstance(X, (pd.DataFrame, np.ndarray, list)):
            raise AlgorithmError("类型错误")

        if isinstance(X, pd.DataFrame):
            X_val = np.mat(X.values)
        elif isinstance(X, (np.ndarray, list)):
            X_val = np.mat(X)

        if y is not None:
            if isinstance(y, (pd.DataFrame, pd.Series)):
                y_val = np.mat(y.values)
            elif isinstance(X, (np.ndarray, list)):
                y_val = np.mat(y)
            return X_val, y_val
        return X_val

    def add_constant(self, X):
        X = self._validate_data(X)
        one_vector = np.ones((X.shape[0], 1))
        return np.hstack((one_vector, X))

    def _decision_function(self, X):
        """
        计算预测值
        :param X:
        :return:
        """
        return X.dot(self.coef_) + self.intercept_

    def predict(self, X):
        X = self._validate_data(X)
        return self._decision_function(X)

    def _set_intercept(self, theta):
        """Set the intercept_"""
        self.intercept_ = theta[0]

    def _set_coef(self, theta):
        self.coef_ = theta[1:]

    def _set_multi_model(self, model):
        self.multi_model_.append(model)


class RegressorMixin:
    def score(self, X: Union[pd.DataFrame, np.ndarray], y: Union[pd.Series, np.ndarray], *args) -> pd.DataFrame:
        """
        获取commom回归分析结果
        :param X:
        :param y:
        :return: 系数，标准误，标准化系数，t，p，R2，调整R2，F
        """
        # 原始值
        X, y = self._validate_data(X, y)
        n, k = X.shape
        # 可能需要的值
        y_pred = self.predict(X)
        x_add = self.add_constant(X)
        intercept_ = self.intercept_
        coef_ = self.coef_
        theta = np.append(intercept_, coef_)
        # 结果
        result_bse = pd.Series(bse(x_add, y, y_pred, n, k), name='标准误')
        result_std_coef = pd.Series(std_coef(coef_, X, y), name='标准化系数')
        result_t = pd.Series(t(theta, x_add, y, y_pred, n, k), name='t')
        result_t_p = pd.Series(t_p(theta, x_add, result_t, n, k), name='p')
        result_r_square = pd.Series(r_square(y, y_pred), name='R2')
        result_adjust_r_square = pd.Series(adjust_r_square(y, y_pred, n, k), name='调整R2')
        result_f = pd.Series(f(y, y_pred, n, k), name='F')
        result_theta = pd.Series(theta, name='B')
        score_1 = pd.concat([result_theta,
                             result_bse,
                             result_std_coef,
                             result_t,
                             result_t_p],
                            axis=1)
        score_2 = pd.concat([result_r_square, result_adjust_r_square, result_f], axis=1)
        return score_1, score_2
